from collections import OrderedDict
import os
import re
from typing import List, Dict

from democritus_algorithms import (
    amb,
    depth_first_traverse,
    breadth_first_traverse,
    genetic_algorithm_best_mutation_function,
    genetic_algorithm_run,
)
from democritus_dicts import dict_values, dict_keys
from democritus_python_data import python_ast_parse

TEST_DATA_1 = {'a': {'b': 1}, 'c': 2, 'd': 3, 'e': {'f': 4}}
TEST_DATA_2 = '''def test():
    """."""
    x = 10
    if x < 20:
        if x < 10:
            print('here')
        return x
    elif x < 30:
        return x
    else:
        x = x * 10
'''


def _get_children_1(data):
    if isinstance(data, dict):
        return dict_values(data)
    else:
        return None


def _get_children_2(data):
    children = []
    if hasattr(data, 'body'):
        children.extend(data.body)
    if hasattr(data, 'orelse'):
        children.extend(data.orelse)
    return children


def _collect_items_2(data):
    return data.__class__.__name__


def test_amb_1():
    """These examples are taken from https://www.rosettacode.org/wiki/Amb#Python."""
    x = (1, 2, 3)
    y = (7, 6, 4, 5)
    assert list(amb(lambda x, y: x * y == 8, x, y)) == [(2, 4)]

    def match(x, y, z):
        return x * x + y * y == z * z

    data_range = range(1, 11)
    assert list(amb(match, data_range, data_range, data_range)) == [(3, 4, 5), (4, 3, 5), (6, 8, 10), (8, 6, 10)]

    l1 = ["the", "that", "a"]
    l2 = ["frog", "elephant", "thing"]
    l3 = ["walked", "treaded", "grows"]
    l4 = ["slowly", "quickly"]

    def is_match(a, b, c, d):
        return a[-1] == b[0] and b[-1] == c[0] and c[-1] == d[0]

    assert list(amb(is_match, l1, l2, l3, l4)) == [('that', 'thing', 'grows', 'slowly')]


def test_depth_first_traverse_1():
    assert list(depth_first_traverse(TEST_DATA_1, _get_children_1)) == [1, 2, 3, 4]

    result = depth_first_traverse(python_ast_parse(TEST_DATA_2), _get_children_2)
    class_names = [i.__class__.__name__ for i in result]
    assert class_names == ['Expr', 'Assign', 'Expr', 'Return', 'Return', 'Assign']


def test_depth_first_traverse_2():
    # this tests a situation where the function to get children and the collect_items_function (_get_children_1) will return None when given a list
    assert list(depth_first_traverse([1, 2, 3], _get_children_1, collect_items_function=_get_children_1)) == [None]


def test_depth_first_traverse_with_collection_function():
    result = tuple(
        depth_first_traverse(python_ast_parse(TEST_DATA_2), _get_children_2, collect_items_function=_collect_items_2)
    )

    assert result == ('Module', 'FunctionDef', 'Expr', 'Assign', 'If', 'If', 'Expr', 'Return', 'If', 'Return', 'Assign')


def test_breadth_first_traverse_1():
    assert list(breadth_first_traverse(TEST_DATA_1, _get_children_1)) == [2, 3, 1, 4]


def test_breadth_first_traverse_with_collection_function():
    result = tuple(
        breadth_first_traverse(python_ast_parse(TEST_DATA_2), _get_children_2, collect_items_function=_collect_items_2)
    )
    assert result == ('Module', 'FunctionDef', 'Expr', 'Assign', 'If', 'If', 'Return', 'If', 'Expr', 'Return', 'Assign')


def test_genetic_algorithm_best_mutation_function_1():
    def scoring_func(item: str) -> int:
        return item.count('1')

    def mutation_func_1(item: str) -> str:
        item = re.sub('0', '1', item, count=1)
        return item

    def mutation_func_2(item: str) -> str:
        item = re.sub('1', '0', item, count=1)
        return item

    starting_items = ['110', '010', '000', '111', '010']
    mutation_funcs = [mutation_func_1, mutation_func_2]
    result = genetic_algorithm_best_mutation_function(starting_items, 100, scoring_func, mutation_funcs)
    assert result == mutation_func_1


def test_genetic_algorithm_run_1():
    def scoring_func(item: str) -> int:
        return item.count('1')

    def selection_func(data: Dict[str, int]) -> List[str]:
        return dict_keys(data)[:3]

    def mutation_func(items: List[str]) -> List[str]:
        for i in items:
            yield re.sub('0', '10', i, count=1)

    starting_items = ['110', '010', '000', '111', '010']

    result = genetic_algorithm_run(starting_items, scoring_func, selection_func, mutation_func, 100)
    assert result == OrderedDict(
        [
            (
                '111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111110',
                101,
            ),
            (
                '111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111111010',
                100,
            ),
            ('111', 3),
        ]
    )



from typing import Any, Callable, Dict, Iterable, List, Optional, Union

from democritus_dicts import dict_sort_by_keys, dict_sort_by_values, dict_values, dict_keys


def amb(validation_function: Callable[..., bool], *args: Any) -> Iterable[Any]:
    import itertools

    products = itertools.product(*args)
    for product in products:
        if validation_function(*product):
            yield product


def depth_first_traverse(
    data: Any,
    get_children_function: Callable[[Any], Optional[Iterable]],
    *,
    collect_items_function: Optional[Callable[[Any], Any]] = None
) -> Iterable[Any]:
    """Traverse the data in a depth-first manner. The get_children_function specifies how children will be identified from each node of the data. The collect_items_function, if provided, allows you to collect items from the data by returning them from the collect_items_function."""
    children = get_children_function(data)

    if collect_items_function:
        yield collect_items_function(data)

    if children:
        for child in children:
            yield from depth_first_traverse(child, get_children_function, collect_items_function=collect_items_function)
    elif not children and not collect_items_function:
        yield data


def breadth_first_traverse(
    data: Any,
    get_children_function: Callable[[Any], Optional[Iterable]],
    *,
    collect_items_function: Optional[Callable[[Any], Any]] = None
) -> Iterable[Any]:
    """Traverse the data in a breadth-first manner. The get_children_function specifies how children will be identified from each node of the data. The collect_items_function, if provided, allows you to collect items from the data by returning them from the collect_items_function."""
    container = [data]
    while len(container) > 0:
        node = container.pop(0)
        if collect_items_function:
            yield collect_items_function(node)

        children = get_children_function(node)
        if children:
            container.extend(children)
        elif not children and not collect_items_function:
            yield node


def genetic_algorithm_run(
    data: Iterable[Any],
    scoring_function: Callable[[Any], Union[int, float]],
    selection_function: Callable[[Dict[Any, Union[int, float]]], Iterable[Any]],
    mutation_function: Callable[[Iterable[Any]], Iterable[Any]],
    max_epochs: int,
) -> Dict[Any, Union[int, float]]:
    scored_data = {}
    for __ in range(max_epochs):
        scored_data = dict_sort_by_values({i: scoring_function(i) for i in data}, reverse=True)
        selected_values = selection_function(scored_data)
        data = mutation_function(selected_values)
    return scored_data


def genetic_algorithm_best_mutation_function(
    starting_values: Iterable[Any],
    generations: int,
    scoring_function: Callable[[Any], Union[int, float]],
    mutation_functions: List[Callable[[Any], Any]],
):
    """Find the best mutation function which produces values from the starting values that score the highest (as measured by the scoring_function) after generations."""
    import statistics

    results = {}

    # TODO: break this out into a genetic_algorithm function
    for mutation_function in mutation_functions:
        values = starting_values
        for generation in range(generations):
            scores = {i: scoring_function(i) for i in values}
            scores = dict_sort_by_keys(scores)
            # todo: keep track of the score data
            values = [mutation_function(val) for val in values]
        final_scores = scores
        average_score = statistics.mean(dict_values(final_scores))
        results[mutation_function] = average_score

    best_mutation_function = dict_keys(dict_sort_by_values(results, reverse=True))[0]
    return best_mutation_function
